﻿using System;
using System.Linq;
using System.Collections.Generic;
using System.Text;
using Startup.Project.Classes;

namespace Startup.Project.Algorithm
{
	public abstract class AlgorithmAction
	{
		public abstract LocalField Result { get; }
		internal abstract string GenerateCodeLine(IList<INamedItem> existedItems);
	}

	public class CreateAction : AlgorithmAction
	{
		private readonly ConstructorDefinition _constructor;
		private readonly FieldBase[] _constructorParameters;
		private readonly LocalField _result;

		public CreateAction(ConstructorDefinition constructor, params FieldBase[] constructorParameters)
		{
			if (constructor == null) 
				throw new ArgumentNullException("constructor");

			_constructor = constructor;
			_constructorParameters = constructorParameters;

			_result = new LocalField(constructor.ReturnedType);
		}

		public override LocalField Result
		{
			get { return _result; }
		}

		internal override string GenerateCodeLine(IList<INamedItem> existedItems)
		{
			_result.UpdateName(existedItems);

			StringBuilder codeLine = new StringBuilder();

			codeLine.AppendFormat("{0} {1} = new {2}(", _result.Type.Name, _result.Name, _constructor.Name);

			for (int index = 0; index < _constructorParameters.Length; index++)
			{
				FieldBase parameter = _constructorParameters[index];

				if (index == 0)
					codeLine.Append(parameter.Name);
				else codeLine.AppendFormat(", {0}", parameter.Name);
			}

			codeLine.Append(");");

			return codeLine.ToString();
		}
	}

	public class GetPropertyValueAction : AlgorithmAction
	{
		private readonly FieldBase _propertyOwner;
		private readonly ReadonlyPropertyDefinition _property;
		private readonly LocalField _result;

		public GetPropertyValueAction(FieldBase propertyOwner, ReadonlyPropertyDefinition property)
		{
			_propertyOwner = propertyOwner;
			_property = property;

			_result = new LocalField(property.ReturnedType);
		}

		public override LocalField Result
		{
			get { return _result; }
		}

		internal override string GenerateCodeLine(IList<INamedItem> existedItems)
		{
			_result.UpdateName(existedItems);

			return string.Format("{0} {1} = {2}.{3};", _result.Type.Name, _result.Name, _propertyOwner.Name, _property.Name);
		}
	}

	public class GetFieldValueAction : AlgorithmAction
	{
		private readonly FieldBase _propertyOwner;
		private readonly FieldBase _propertyTarget;
		private readonly LocalField _result;

		public GetFieldValueAction(FieldBase propertyOwner, FieldBase propertyTarget)
		{
			_propertyOwner = propertyOwner;
			_propertyTarget = propertyTarget;

			_result = new LocalField(propertyTarget.Type);
		}

		public override LocalField Result
		{
			get { return _result; }
		}

		internal override string GenerateCodeLine(IList<INamedItem> existedItems)
		{
			_result.UpdateName(existedItems);

			return string.Format("{0} {1} = {2}.{3};", _result.Type.Name, _result.Name, _propertyOwner.Name, _propertyTarget.Name);
		}
	}

	public class SetPropertyValueAction : AlgorithmAction
	{
		private readonly FieldBase _propertyOwner;
		private readonly PropertyDefinition _property;
		private readonly FieldBase _value;

		public SetPropertyValueAction(FieldBase propertyOwner, PropertyDefinition property, FieldBase value)
		{
			_propertyOwner = propertyOwner;
			_property = property;
			_value = value;
		}

		public override LocalField Result
		{
			get { return null; }
		}

		internal override string GenerateCodeLine(IList<INamedItem> existedItems)
		{
			return string.Format("{0}.{1} = {2};", _propertyOwner.Name, _property.Name, _value.Name);
		}
	}

	public class SetFieldValueAction : AlgorithmAction
	{
		private readonly FieldBase _fieldOwner;
		private readonly FieldBase _field;
		private readonly FieldBase _value;

		public SetFieldValueAction(FieldBase fieldOwner, FieldBase field, FieldBase value)
		{
			_fieldOwner = fieldOwner;
			_field = field;
			_value = value;
		}

		public override LocalField Result
		{
			get { return null; }
		}

		internal override string GenerateCodeLine(IList<INamedItem> existedItems)
		{
			return string.Format("{0}.{1} = {2};", _fieldOwner.Name, _field.Name, _value.Name);
		}
	}

	public class CallMethodAction : AlgorithmAction
	{
		private readonly FieldBase _methodOwner;
		private readonly MethodDefinition _method;
		private readonly FieldBase[] _methodParameters;
		private readonly LocalField _result;

		public CallMethodAction(FieldBase methodOwner, MethodDefinition method, params FieldBase[] methodParameters)
		{
			_methodOwner = methodOwner;
			_method = method;
			_methodParameters = methodParameters ?? new FieldBase[0];

			_result = _method.ReturnedType == null ? null : new LocalField(method.ReturnedType);
		}

		public override LocalField Result
		{
			get { return _result; }
		}

		internal override string GenerateCodeLine(IList<INamedItem> existedItems)
		{
			if (_result != null)
				_result.UpdateName(existedItems);

			StringBuilder codeLine = new StringBuilder();

			if (_result != null)
				codeLine.AppendFormat("{0} {1} = ", _result.Type.Name, _result.Name);

			codeLine.AppendFormat("{0}.{1}(", _methodOwner.Name, _method.Name);

			for (int index = 0; index < _methodParameters.Length; index++)
			{
				FieldBase parameter = _methodParameters[index];

				if (index == 0)
					codeLine.Append(parameter.Name);
				else codeLine.AppendFormat(", {0}", parameter.Name);
			}

			codeLine.Append(");");

			return codeLine.ToString();
		}
	}

	public class ReturnAction : AlgorithmAction
	{
		private readonly FieldBase _returnedValue;

		public ReturnAction(FieldBase returnedValue)
		{
			_returnedValue = returnedValue;
		}

		public override LocalField Result
		{
			get { return null; }
		}

		internal override string GenerateCodeLine(IList<INamedItem> existedItems)
		{
			return string.Format("return {0};", _returnedValue.Name);
		}
	}

	public class CustomAction : AlgorithmAction
	{
		private readonly string _format;
		private readonly INamedItem[] _args;

		public CustomAction(string format, params INamedItem[] args)
		{
			_format = format;
			_args = args;
		}

		public override LocalField Result
		{
			get { return null; }
		}

		internal override string GenerateCodeLine(IList<INamedItem> existedItems)
		{
			return string.Format(_format, _args.Select(x => (object)x.Name).ToArray());
		}
	}
}
