﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Startup.Base.Utils;
using Startup.Project.Classes;
using Startup.Project.Utils;

namespace Startup.Project.Algorithm
{
	public abstract class AlgorithmDefinition
	{
		private readonly List<AlgorithmAction> _actions = new List<AlgorithmAction>();

		public abstract string Name { get; }
		public abstract ClassBase ReturnedType { get; }

		public void AddCodeLine(AlgorithmAction action)
		{
			_actions.Add(action);
		}

		internal string GenerateCode(IList<INamedItem> existedItems)
		{
			IList<INamedItem> actualExistedItems = CreateCopyForCurrentEnvironment(existedItems);

			TabStringBuilder code = new TabStringBuilder();
			code.Tab().Tab().AddLine(GetAlgorithmCodeHeader());
			code.Tab().Tab().AddLine("{");

			for (int index = 0; index < _actions.Count; index++)
			{
				AlgorithmAction action = _actions[index];
				code.Tab().Tab().Tab().AddLine(action.GenerateCodeLine(actualExistedItems));

				if (action.Result != null)
					actualExistedItems.Add(action.Result);
			}

			code.Tab().Tab().AddLine("}");

			return code.ToString();
		}

		internal virtual IList<INamedItem> CreateCopyForCurrentEnvironment(IList<INamedItem> existedItems)
		{
			return new List<INamedItem>(existedItems);
		}

		protected abstract string GetAlgorithmCodeHeader();

		public virtual IEnumerable<string> NeededNamespaces
		{
			get
			{
				foreach (AlgorithmAction action in _actions)
				{
					if (action.Result == null)
						continue;

					yield return action.Result.Type.Namespace;

					if (action.Result.Type is CollectionStandartClass)
						yield return ((CollectionStandartClass)action.Result.Type).TemplateType.Namespace;
				}
			}
		}
	}

	public class ConstructorDefinition : AlgorithmDefinition
	{
		private readonly ClassBase _constructorOwner;
		private readonly MethodParameter[] _constructorParameters;
		private readonly Dictionary<MethodParameter, Field> _parametersBindings = new Dictionary<MethodParameter, Field>(); 

		public ConstructorDefinition(ClassBase constructorOwner, params MethodParameter[] constructorParameters)
		{
			if (constructorOwner == null) 
				throw new ArgumentNullException("constructorOwner");

			_constructorOwner = constructorOwner;
			_constructorParameters = constructorParameters ?? new MethodParameter[0];
		}

		public MethodParameter[] ConstructorParameters
		{
			get { return _constructorParameters; }
		}

		public override string Name
		{
			get { return _constructorOwner.Name; }
		}

		public override ClassBase ReturnedType
		{
			get { return _constructorOwner; }
		}

		public override IEnumerable<string> NeededNamespaces
		{
			get
			{
				foreach (string neededNamespace in base.NeededNamespaces)
					yield return neededNamespace;

				foreach (MethodParameter param in _constructorParameters)
				{
					yield return param.Type.Namespace;

					if (param.Type is CollectionStandartClass)
						yield return ((CollectionStandartClass)param.Type).TemplateType.Namespace;
				}
			}
		}

		internal override IList<INamedItem> CreateCopyForCurrentEnvironment(IList<INamedItem> existedItems)
		{
			IList<INamedItem> copy = new List<INamedItem>(existedItems);

			for (int index = 0; index < _constructorParameters.Length; index++)
				copy.Add(_constructorParameters[index]);

			return copy;
		}

		protected override string GetAlgorithmCodeHeader()
		{
			StringBuilder codeLine = new StringBuilder();

			codeLine.AppendFormat("public {0}(", CSharpNameHelper.GetPropertyName(Name));

			if (_constructorParameters != null)
			{
				for (int index = 0; index < _constructorParameters.Length; index++)
				{
					FieldBase parameter = _constructorParameters[index];

					if (index == 0)
						codeLine.AppendFormat("{0} {1}", parameter.Type, parameter.Name);
					else codeLine.AppendFormat(", {0} {1}", parameter.Type, parameter.Name);
				}
			}

			codeLine.Append(")");

			return codeLine.ToString();
		}

		internal void RegisterParameterBinding(MethodParameter parameter, Field backgroundField)
		{
			if (parameter == null)
				throw new ArgumentNullException("parameter");
			if (backgroundField == null)
				throw new ArgumentNullException("backgroundField");

			_parametersBindings.Add(parameter, backgroundField);
		}

		public Field GetBindedField(MethodParameter parameter)
		{
			if (parameter == null) 
				throw new ArgumentNullException("parameter");

			Field target;

			if (_parametersBindings.TryGetValue(parameter, out target))
				return target;

			return null;
		}
	}

	public class MethodDefinition : AlgorithmDefinition
	{
		private readonly ClassBase _returnedType;
		private readonly string _name;
		private readonly MethodParameter[] _methodParameters;

		public MethodDefinition(string name)
			: this(null, name)
		{}

		public MethodDefinition(ClassBase returnedType, string name, params MethodParameter[] methodParameters)
		{
			_returnedType = returnedType;
			_name = name;
			_methodParameters = methodParameters ?? new MethodParameter[0];
		}

		public MethodParameter[] MethodParameters
		{
			get { return _methodParameters; }
		}

		public override string Name
		{
			get { return _name; }
		}

		public override ClassBase ReturnedType
		{
			get { return _returnedType; }
		}

		public override IEnumerable<string> NeededNamespaces
		{
			get
			{
				foreach (string neededNamespace in base.NeededNamespaces)
					yield return neededNamespace;

				foreach (MethodParameter param in _methodParameters)
				{
					yield return param.Type.Namespace;

					if (param.Type is CollectionStandartClass)
						yield return ((CollectionStandartClass) param.Type).TemplateType.Namespace;
				}
			}
		}

		protected override string GetAlgorithmCodeHeader()
		{
			StringBuilder codeLine = new StringBuilder();

			codeLine.AppendFormat("public {0} {1}(", ReturnedType == null ? "void" : ReturnedType.Name, CSharpNameHelper.GetPropertyName(Name));

			if (_methodParameters != null)
			{
				for (int index = 0; index < _methodParameters.Length; index++)
				{
					FieldBase parameter = _methodParameters[index];

					if (index == 0)
						codeLine.AppendFormat("{0} {1}", parameter.Type, parameter.Name);
					else codeLine.AppendFormat(", {0} {1}", parameter.Type, parameter.Name);
				}
			}

			codeLine.Append(")");

			return codeLine.ToString();
		}
	}

	public class ReadonlyPropertyDefinition : INamedItem
	{
		private readonly ClassBase _returnedType;
		private readonly string _name;
		private readonly bool _isStatic;

		private readonly List<AlgorithmAction> _getActions = new List<AlgorithmAction>();

		public ReadonlyPropertyDefinition(ClassBase returnedType, string name, bool isStatic)
		{
			if (returnedType == null) 
				throw new ArgumentNullException("returnedType");
			if (string.IsNullOrEmpty(name))
				throw new ArgumentNullException("name");

			_returnedType = returnedType;
			_name = name;
			_isStatic = isStatic;
		}

		public string Name
		{
			get { return _name; }
		}

		public bool IsStatic
		{
			get { return _isStatic; }
		}

		public ClassBase ReturnedType
		{
			get { return _returnedType; }
		}

		public IEnumerable<string> NeededNamespaces
		{
			get
			{
				foreach (AlgorithmAction action in _getActions)
				{
					if (action.Result == null)
						continue;

					yield return action.Result.Type.Namespace;

					if (action.Result.Type is CollectionStandartClass)
						yield return ((CollectionStandartClass) action.Result.Type).TemplateType.Namespace;
				}
			}
		}

		public void AddCodeLineToGetter(AlgorithmAction action)
		{
			_getActions.Add(action);
		}

		internal string GenerateCodeForGetter(IList<INamedItem> existedItems)
		{
			TabStringBuilder code = new TabStringBuilder();

			// getter
			code.Tab().Tab().Tab().AddLine("get");
			code.Tab().Tab().Tab().AddLine("{");

			IList<INamedItem> actualExistedItems = new List<INamedItem>(existedItems);
			for (int index = 0; index < _getActions.Count; index++)
			{
				AlgorithmAction action = _getActions[index];
				code.Tab().Tab().Tab().Tab().AddLine(action.GenerateCodeLine(actualExistedItems));

				if (action.Result != null)
					actualExistedItems.Add(action.Result);
			}

			code.Tab().Tab().Tab().AddLine("}");

			return code.ToString();
		}

		internal virtual string GenerateCode(IList<INamedItem> existedItems)
		{
			TabStringBuilder code = new TabStringBuilder();
			code.Tab().Tab().AddLine("public {0}{1} {2}", _isStatic ? "static " : string.Empty, _returnedType.Name, _name);
			code.Tab().Tab().AddLine("{");

			code.AddText(GenerateCodeForGetter(existedItems));
			
			code.Tab().Tab().AddLine("}");

			return code.ToString();
		}
	}

	public class PropertyDefinition : ReadonlyPropertyDefinition
	{
		private readonly List<AlgorithmAction> _setActions = new List<AlgorithmAction>();
		
		public PropertyDefinition(ClassBase returnedType, string name, bool isStatic)
			: base(returnedType, name, isStatic)
		{ }

		public void AddCodeLineToSetter(AlgorithmAction action)
		{
			_setActions.Add(action);
		}

		internal override string GenerateCode(IList<INamedItem> existedItems)
		{
			TabStringBuilder code = new TabStringBuilder();
			code.Tab().Tab().AddLine("public {0}{1} {2}", IsStatic ? "static " : string.Empty, ReturnedType.Name, Name);
			code.Tab().Tab().AddLine("{");

			// getter
			code.AddText(GenerateCodeForGetter(existedItems));
			
			// setter
			code.Tab().Tab().Tab().AddLine("set");
			code.Tab().Tab().Tab().AddLine("{");

			List<INamedItem> actualExistedItems = new List<INamedItem>(existedItems);
			actualExistedItems.Add(new ValueField(ReturnedType));
			for (int index = 0; index < _setActions.Count; index++)
			{
				AlgorithmAction action = _setActions[index];
				code.Tab().Tab().Tab().Tab().AddLine(action.GenerateCodeLine(actualExistedItems));

				if (action.Result != null)
					actualExistedItems.Add(action.Result);
			}

			code.Tab().Tab().Tab().AddLine("}");

			// close root
			code.Tab().Tab().AddLine("}");

			return code.ToString();
		}
	}
}
