﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using Startup.Base.Utils;
using Startup.Project.Algorithm;
using Startup.Project.Interfaces;
using Startup.Project.Utils;

namespace Startup.Project.Classes
{
	public class CustomClass : ClassBase
	{
		internal const string Extension = ".cs";

		private readonly bool _isPartOfXaml;
		private readonly ClassBase _baseClass;
		
		private readonly List<Field> _fields = new List<Field>();
		private readonly List<ConstructorDefinition> _constructors = new List<ConstructorDefinition>();
		private readonly List<MethodDefinition> _methods = new List<MethodDefinition>();
		private readonly List<ReadonlyPropertyDefinition> _properties = new List<ReadonlyPropertyDefinition>();
		
		internal CustomClass(string name, string namespaceName, bool isPartOfXaml)
			: this(name, namespaceName, isPartOfXaml, new StandartClass(typeof(Object).Name, typeof(Object).Namespace))
		{}

		internal CustomClass(string name, string namespaceName, bool isPartOfXaml, ClassBase baseClass)
			: base(name, namespaceName)
		{
			_isPartOfXaml = isPartOfXaml;
			_baseClass = baseClass;
		}

		public ConstructorDefinition DefaultConstructor
		{
			get
			{
				if (_constructors.Count == 0)
					return new ConstructorDefinition(this);

				return _constructors.FirstOrDefault(x => x.ConstructorParameters.Length == 0);
			}
		}

		internal virtual string NameWithExtension
		{
			get
			{
				if (_isPartOfXaml)
					return Name + ViewBase.Extension + Extension;

				return Name + Extension;
			}
		}

		public bool IsPartOfXaml
		{
			get { return _isPartOfXaml; }
		}

		public ReadonlyPropertyDefinition AddAutoProperty(string name, ClassBase type)
		{
			if (string.IsNullOrEmpty(name))
				throw new ArgumentNullException("name");
			if (type == null)
				throw new ArgumentNullException("type");

			string fieldName = CSharpNameHelper.GetFieldName(name);
			Field field = AddField(fieldName, type, false);

			return AddProperty(field, name, type);
		}

		public FieldBase This
		{
			get { return new ThisField(this); }
		}

		public List<ReadonlyPropertyDefinition> Properties
		{
			get { return _properties; }
		}

		public List<Field> Fields
		{
			get { return _fields; }
		}

		public List<ConstructorDefinition> Constructors
		{
			get { return _constructors; }
		}

		public Field AddField(string name, ClassBase type, bool isReadonly)
		{
			return AddField(name, type, isReadonly, false);
		}

		public Field AddField(string name, ClassBase type, bool isReadonly, bool createByDefaultConstructor)
		{
			if (type == null) 
				throw new ArgumentNullException("type");
			if (string.IsNullOrEmpty(name))
				throw new ArgumentNullException("name");

			Field field = new Field(name, type, isReadonly, createByDefaultConstructor);
			_fields.Add(field);

			return field;
		}

		public MethodDefinition AddMethod(string name, ClassBase returntedType, params MethodParameter[] parametrs)
		{
			MethodDefinition method = new MethodDefinition(returntedType, name, parametrs);
			_methods.Add(method);

			return method;
		}
		
		public ConstructorDefinition AddConstructor(params MethodParameter[] parametrs)
		{
			ConstructorDefinition constructor = new ConstructorDefinition(this, parametrs);
			
			for (int index = 0; index < parametrs.Length; index++)
			{
				MethodParameter parameter = parametrs[index];

				string backgroundFieldName = CSharpNameHelper.GetFieldName(parameter.Name);

				Field backgroundField = 
					_fields.FirstOrDefault(x => x.Name == backgroundFieldName && x.Type == parameter.Type) ??
					AddField(backgroundFieldName, parameter.Type, false);

				SetFieldValueAction setField = new SetFieldValueAction(This, backgroundField, parameter);
				constructor.AddCodeLine(setField);
				constructor.RegisterParameterBinding(parameter, backgroundField);
			}
			
			_constructors.Add(constructor);
			return constructor;
		}
		
		internal void Save(string classPath)
		{
			List<INamedItem> existedItems = new List<INamedItem>();
			TabStringBuilder code = new TabStringBuilder();

			// namespaces
			string[] usings = 
				_properties.Select(x => x.ReturnedType.Namespace).Concat(
				_properties.SelectMany(x => x.NeededNamespaces)).Concat(
				_constructors.SelectMany(x => x.NeededNamespaces)).Concat(
				_methods.SelectMany(x => x.NeededNamespaces)).Concat(
				_methods.Where(x => x.ReturnedType != null).Select(x => x.ReturnedType.Namespace)).Concat(
                _fields.Select(x => x.Type.Namespace)).
				Distinct().ToArray();

			Array.ForEach(usings, usedNamespace => code.AddLine("using {0};", usedNamespace));
			code.AddLine("using {0};", _baseClass.Namespace);
			code.AddLine("using Startup.Base.Commands;");
			code.AddLine("using System.Xml.Serialization;");
			code.AddLine("using Microsoft.Win32;");
			
			// class
			code.AddLine("namespace {0}", Namespace);
			code.AddLine("{");

			code.Tab().AddLine("public {0}class {1} : {2}", _isPartOfXaml ? "partial " : string.Empty, Name, _baseClass.Name);
			code.Tab().AddLine("{");

            // readonly fields
			_fields.ForEach(x =>
			{
				code.Tab().Tab().AddLine(x.GenerateCode());
				existedItems.Add(x);
			});
			code.AddLine();

			// properties
			_properties.ForEach(x =>
			{
				code.AddLine(x.GenerateCode(existedItems));
				existedItems.Add(x);
			});

			// constructor header
			_constructors.ForEach(x => code.AddLine(x.GenerateCode(existedItems)));
			
			// methods
			_methods.ForEach(x => code.AddLine(x.GenerateCode(existedItems)));

			code.Tab().AddLine("}");
			code.AddLine("}");

			File.WriteAllText(classPath, code.ToString());
		}
		
		public ReadonlyPropertyDefinition AddReadonlyProperty(Field backgroundField)
	    {
	        if (backgroundField == null)
                throw new ArgumentNullException("backgroundField");

			string propertyName = CSharpNameHelper.GetPropertyName(backgroundField.Name);
			ReadonlyPropertyDefinition property = new ReadonlyPropertyDefinition(backgroundField.Type, propertyName, false);

			ReturnAction returnAction = new ReturnAction(backgroundField);
			property.AddCodeLineToGetter(returnAction);

			_properties.Add(property);
			return property;
	    }

		public ReadonlyPropertyDefinition AddReadonlyProperty(string path, string propertyName, ClassBase propertyType)
	    {
	        if (propertyType == null)
                throw new ArgumentNullException("propertyType");
	        if (string.IsNullOrEmpty(propertyName)) 
                throw new ArgumentNullException("propertyName");
            if (string.IsNullOrEmpty(path))
                throw new ArgumentNullException("path");

			Field fieldImmitation = new Field(path, propertyType, false, false);
			ReadonlyPropertyDefinition property = new ReadonlyPropertyDefinition(propertyType, propertyName, false);

			ReturnAction returnAction = new ReturnAction(fieldImmitation);
			property.AddCodeLineToGetter(returnAction);

			_properties.Add(property);
			return property;
	    }

		public void AddProperty(ReadonlyPropertyDefinition property)
		{
			if (property == null) 
				throw new ArgumentNullException("property");

			_properties.Add(property);
		}

		public ReadonlyPropertyDefinition AddProperty(string path, string propertyName, ClassBase propertyType)
	    {
	    	if (string.IsNullOrEmpty(path)) 
				throw new ArgumentNullException("path");
	    	if (string.IsNullOrEmpty(propertyName))
				throw new ArgumentNullException("propertyName");
	    	if (propertyType == null)
				throw new ArgumentNullException("propertyType");
	    	
			Field field = new Field(path, propertyType, false, false);
	    	return AddProperty(field, propertyName, propertyType);
	    }

		public PropertyDefinition AddProperty(Field field, string propertyName, ClassBase propertyType)
	    {
	        if (propertyType == null)
                throw new ArgumentNullException("propertyType");
	        if (string.IsNullOrEmpty(propertyName)) 
                throw new ArgumentNullException("propertyName");
			if (field == null)
				throw new ArgumentNullException("field");

			PropertyDefinition property = new PropertyDefinition(propertyType, propertyName, false);

			// setter
			ReturnAction returnAction = new ReturnAction(field);
			property.AddCodeLineToGetter(returnAction);

			// setter
			SetFieldValueAction setAction = new SetFieldValueAction(new ThisField(this), field, new ValueField(propertyType));
			property.AddCodeLineToSetter(setAction);

            _properties.Add(property);
	    	return property;
	    }
	}
}
