﻿using System.Linq;
using System.Collections.Generic;
using System.Globalization;
using System.Text;
using Startup.Project.Utils;

namespace Startup.Project.Classes
{
	public abstract class FieldBase : INamedItem
	{
		public abstract string Name { get; }
		public abstract ClassBase Type { get; }

		public override string ToString()
		{
			return string.Format("{0} {1}", Type, Name);
		}
	}

	public class ThisField : FieldBase
	{
		private readonly ClassBase _classItem;

		public ThisField(ClassBase classItem)
		{
			_classItem = classItem;
		}

		public override string Name
		{
			get { return "this"; }
		}

		public override ClassBase Type
		{
			get { return _classItem; }
		}
	}

	public class ValueField : FieldBase
	{
		private readonly ClassBase _classItem;

		public ValueField(ClassBase classItem)
		{
			_classItem = classItem;
		}

		public override string Name
		{
			get { return "value"; }
		}

		public override ClassBase Type
		{
			get { return _classItem; }
		}
	}

	public class Field : FieldBase
    {
        private readonly string _name;
        private readonly ClassBase _type;
        private readonly bool _isReadonly;
		private readonly bool _createByDefaultConstructor;

		public Field(string name, ClassBase type, bool isReadonly, bool createByDefaultConstructor)
        {
            _name = name;
            _type = type;
            _isReadonly = isReadonly;
        	_createByDefaultConstructor = createByDefaultConstructor;
        }

        public override string Name
        {
            get { return _name; }
        }

        public override ClassBase Type
        {
            get { return _type; }
        }

        public bool IsReadonly
        {
            get { return _isReadonly; }
        }

		internal string GenerateCode()
		{
			StringBuilder builder = new StringBuilder();

			builder.Append("private ");

			if (_isReadonly)
				builder.Append("readonly ");

			builder.AppendFormat("{0} {1}", Type.Name, Name);

			if (_createByDefaultConstructor)
				builder.AppendFormat(" = new {0}()", Type.Name);

			builder.Append(";");

			return builder.ToString();
		}
    }

	public class LocalField : FieldBase
	{
		private readonly ClassBase _type;

		private string _name;
        
		public LocalField(ClassBase type)
        {
            _type = type;
        }

        public override string Name
        {
            get { return _name; }
        }

        public override ClassBase Type
        {
            get { return _type; }
        }

		internal void UpdateName(IList<INamedItem> existedItems)
		{
			// check for empty existedItems collection
			string nameBase = CSharpNameHelper.GetLocalFieldName(_type.Name);

			if (nameBase == "string" ||
				nameBase == "int" ||
				nameBase == "bool")
			{
				nameBase = "temp";
			}

			if (existedItems == null)
			{
				_name = nameBase;
				return;
			}

			// remove this item if it's already situated there
			for (int index = existedItems.Count - 1; index >= 0; index--)
			{
				INamedItem item = existedItems[index];

				if (item.Name == _name)
					existedItems.RemoveAt(index);
			}

			// register this item
			string tempName = nameBase;
			for (int index = 1;; index++)
			{
				if (existedItems.All(x => x.Name != tempName))
				{
					_name = tempName;
					return;
				}

				tempName = string.Format("{0}{1}", nameBase, index);
			}
		}
	}
}
