﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using Startup.Base.Commands;
using Startup.Base.Utils;
using Startup.Project.Algorithm;
using Startup.Project.Classes;
using Startup.Project.Utils;

namespace Startup.Project.Interfaces
{
	public class ViewContainer : ViewBase
	{
		public ViewContainer(string name, string parentNamespace, CustomClass viewModelClass, CustomClass modelClass, string relativeParentPath)
			: base(name, parentNamespace, viewModelClass, modelClass, relativeParentPath)
		{ }

		internal override void Save(string filePath)
		{
			if (string.IsNullOrEmpty(filePath))
				throw new ArgumentNullException("filePath");

			UiGridCombinerHelper gridContext = new UiGridCombinerHelper();
			string appXamlContent = SaveToString(gridContext);
			File.WriteAllText(filePath, appXamlContent);
		}

		internal override string SaveToString(UiGridCombinerHelper gridContext)
		{
			TabStringBuilder appXamlContent = new TabStringBuilder();
			appXamlContent.AddLine("<UserControl x:Class=\"{0}.{1}\"", ViewModelClass.Namespace, Name);
			appXamlContent.Tab().AddLine("xmlns=\"http://schemas.microsoft.com/winfx/2006/xaml/presentation\"");
			appXamlContent.Tab().AddLine("xmlns:x=\"http://schemas.microsoft.com/winfx/2006/xaml\"");
			appXamlContent.Tab().AddLine("xmlns:b=\"http://schemas.microsoft.com/expression/blend/2008\"");
			appXamlContent.Tab().AddLine("xmlns:mc=\"http://schemas.openxmlformats.org/markup-compatibility/2006\"");
			appXamlContent.Tab().AddLine("mc:Ignorable=\"b\"");

			Children.SelectMany(GetAllViewContainers).ToList().ForEach(x => appXamlContent.Tab().AddLine("xmlns:{0}=\"clr-namespace:{1}\"", CSharpNameHelper.GetLocalFieldName(x.Name), x.Namespace));

			appXamlContent.Tab().AddLine("xmlns:ViewModelHome=\"clr-namespace:{0}\"", ViewModelClass.Namespace);
			appXamlContent.Tab().AddLine("b:DataContext=\"{x:Static ViewModelHome:" + ViewModelClass.Name + ".Designer}\"");
			appXamlContent.Tab().AddLine("xmlns:Shell=\"clr-namespace:{0}\"", ParentNamespace);
			appXamlContent.Tab().AddLine("xmlns:xctk=\"http://schemas.xceed.com/wpf/xaml/toolkit\" >");
			
			for (int index = 0; index < Children.Length; index++)
			{
				UiContainer child = Children[index];

				if (child == null)
					continue;

				string childContent = child.SaveToString(gridContext);
				appXamlContent.AddLine(childContent);
			}

			appXamlContent.AddLine("</UserControl>");

			return appXamlContent.ToString();
		}

		internal static List<ViewLinkContainer> GetAllViewContainers(UiContainer target)
		{
			if (target == null)
				return new List<ViewLinkContainer>();

			if (target is ViewLinkContainer)
				return new List<ViewLinkContainer>(new[] { (ViewLinkContainer) target } );

			return target.Children.SelectMany(GetAllViewContainers).ToList();
		}

		public override void AddCommands(List<CommandDefinition> commands)
		{
			commands.ForEach(x =>
			{
				// view model
				MethodDefinition onCommand = ViewModelClass.AddMethod("On" + x.CommandName, null);
				ConstructorDefinition viewModelConstructor = ViewModelClass.Constructors.First();
				Field modelField = viewModelConstructor.GetBindedField(viewModelConstructor.ConstructorParameters.First());
				onCommand.AddCodeLine(new CallMethodAction(modelField, new MethodDefinition("On" + x.CommandName)));
				//onCommand.AddCodeLine(new CallMethodAction(new ThisField(ViewModelClass), new MethodDefinition("RaisePropertyChanged"), new Field('"' + x.TargetProperty + '"', new StandartClass("String", "System"), false, false)));

				ViewModelClass.Properties.ForEach(y => onCommand.AddCodeLine(new CallMethodAction(new ThisField(ViewModelClass), new MethodDefinition("RaisePropertyChanged"), new Field('"' + y.Name + '"', new StandartClass("String", "System"), false, false))));

				StandartClass simpleCommand = new StandartClass(typeof(SimpleCommand).Name, typeof(SimpleCommand).Namespace);
				ReadonlyPropertyDefinition commandProperty = new ReadonlyPropertyDefinition(simpleCommand, x.CommandName, false);
				CreateAction createDelegate = new CreateAction(new ConstructorDefinition(simpleCommand, new MethodParameter("method", simpleCommand)), new Field("On" + x.CommandName, simpleCommand, false, false));
				ReturnAction returnCommand = new ReturnAction(createDelegate.Result);
				commandProperty.AddCodeLineToGetter(createDelegate);
				commandProperty.AddCodeLineToGetter(returnCommand);

				ViewModelClass.AddProperty(commandProperty);

				// model
				MethodDefinition commandBody = ModelClass.AddMethod("On" + x.CommandName, null, null);
				Array.ForEach(x.CommandContent, y => commandBody.AddCodeLine(y));
			});
		}
	}
}
