﻿using System;
using System.Diagnostics.Contracts;
using System.Linq;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using Startup.MVVM;
using Startup.Project.Algorithm;
using Startup.Project.Classes;
using Startup.Project.Content;
using Startup.Project.Interfaces;
using Startup.Project.Utils;
using PropertyPath = Startup.Project.Utils.PropertyPath;
using ViewBase = Startup.Project.Interfaces.ViewBase;

namespace Startup.Project.Project
{
	public class Folder : ISourceFileContainer
	{
		private readonly string _name;
		private readonly string _parentNamespace;
		private readonly string _relativeParentPath;
		private readonly string _namespace;

		private readonly List<CustomClass> _classes = new List<CustomClass>();
		private readonly List<Folder> _folders = new List<Folder>();
		private readonly List<ViewBase> _views = new List<ViewBase>();

		internal Folder(string name, string parentNamespace, string relativeParentPath)
		{
			if (string.IsNullOrEmpty(name))
				throw new ArgumentNullException("name");

			_name = name;
			_parentNamespace = parentNamespace;
			_relativeParentPath = relativeParentPath;

			_namespace = string.IsNullOrEmpty(parentNamespace) ? 
				_name : string.Format("{0}.{1}", parentNamespace, _name);
		}

		internal string Name
		{
			get { return _name; }
		}

		internal List<CustomClass> Classes
		{
			get { return _classes; }
		}

		internal List<Folder> Folders
		{
			get { return _folders; }
		}

		internal string Namespace
		{
			get { return _namespace; }
		}

		internal virtual string RelativePath
		{
			get { return Path.Combine(_relativeParentPath, Name); }
		}

		public CustomClass AddClass(string name)
		{
			return AddClass(name, new StandartClass(typeof (object)));
		}

		public CustomClass AddClass(string name, ClassBase parentClass)
		{
			if (parentClass == null) 
				throw new ArgumentNullException("parentClass");
			if (string.IsNullOrEmpty(name))
				throw new ArgumentNullException("name");

			var instance = new CustomClass(name, Namespace, false, parentClass);
			_classes.Add(instance);

			return instance;
		}

		public ISourceFileContainer AddFolder(string name)
		{
			if (string.IsNullOrEmpty(name))
				throw new ArgumentNullException("name");

			var instance = new Folder(name, Namespace, RelativePath);
			_folders.Add(instance);

			return instance;
		}

		public ViewDefinition AddView(string name, ClassBase sourceClass, ClassContent designerContent)
		{
			if (string.IsNullOrEmpty(name))
				throw new ArgumentException("Name cannot be null.", "name");
			if (sourceClass == null) 
				throw new ArgumentNullException("sourceClass");
			if (designerContent == null) 
				throw new ArgumentNullException("designerContent");

			// create model
			CustomClass model = AddClass(name + "Model");
			MethodParameter modelConstructorInput = new MethodParameter(CSharpNameHelper.GetLocalFieldName(sourceClass.Name), sourceClass);
			ConstructorDefinition contructorModel = model.AddConstructor(modelConstructorInput);
			
			// create view model
			CustomClass viewModel = AddClass(name + "ViewModel", new StandartClass(typeof(ViewModelBase)));
			MethodParameter viewModelConstructorInput = new MethodParameter(CSharpNameHelper.GetLocalFieldName(model.Name), model);
			ConstructorDefinition contructorViewModel = viewModel.AddConstructor(viewModelConstructorInput);

			// create view
			CustomClass backgroundClass = new CustomClass(name + "View", Namespace, true, new StandartClass(typeof(UserControl).Name, typeof(UserControl).Namespace));
			ConstructorDefinition constructor = backgroundClass.AddConstructor();
			constructor.AddCodeLine(new CallMethodAction(new ThisField(backgroundClass), new MethodDefinition("InitializeComponent")));

			ViewContainer view = new ViewContainer(name + "View", _parentNamespace, viewModel, model, RelativePath);

			_classes.Add(backgroundClass);
			_views.Add(view);

			// view model : designer property
			ReadonlyPropertyDefinition designerProperty = new ReadonlyPropertyDefinition(viewModel, "Designer", true);
			viewModel.AddProperty(designerProperty);

			CreateAction createDomain = new CreateAction(sourceClass is CustomClass ?
				((CustomClass)sourceClass).DefaultConstructor : new ConstructorDefinition(sourceClass));
			designerProperty.AddCodeLineToGetter(createDomain);

			for (int index = 0; index < designerContent.Properties.Count; index++)
				AddInitializationActionsFor(designerProperty, createDomain.Result, designerContent.Properties[index]);

			CreateAction createModel = new CreateAction(contructorModel, createDomain.Result);
			designerProperty.AddCodeLineToGetter(createModel);

			CreateAction createViewModel = new CreateAction(contructorViewModel, createModel.Result);
			designerProperty.AddCodeLineToGetter(createViewModel);

			ReturnAction retirnViewModelForDesigner = new ReturnAction(createViewModel.Result);
			designerProperty.AddCodeLineToGetter(retirnViewModelForDesigner);

			return new ViewDefinition(model, viewModel, view);
		}

		public WindowDefinition AddWindow(string name, ClassBase sourceClass, ClassContent designerContent)
		{
			if (string.IsNullOrEmpty(name))
				throw new ArgumentException("Name cannot be null.", "name");
			if (sourceClass == null) 
				throw new ArgumentNullException("sourceClass");
			if (designerContent == null) 
				throw new ArgumentNullException("designerContent");

			// create model
			CustomClass model = AddClass(name + "Model");
			MethodParameter modelConstructorInput = new MethodParameter(CSharpNameHelper.GetLocalFieldName(sourceClass.Name), sourceClass);
			ConstructorDefinition contructorModel = model.AddConstructor(modelConstructorInput);
			
			// create view model
			CustomClass viewModel = AddClass(name + "ViewModel", new StandartClass(typeof(ViewModelBase)));
			MethodParameter viewModelConstructorInput = new MethodParameter(CSharpNameHelper.GetLocalFieldName(model.Name), model);
			ConstructorDefinition contructorViewModel = viewModel.AddConstructor(viewModelConstructorInput);

			// create view model
			CustomClass backgroundClass = new CustomClass(name + "View", Namespace, true, new StandartClass(typeof(Window).Name, typeof(Window).Namespace));
			ConstructorDefinition constructor = backgroundClass.AddConstructor();
			constructor.AddCodeLine(new CallMethodAction(new ThisField(backgroundClass), new MethodDefinition("InitializeComponent")));

			WindowContainer view = new WindowContainer(name + "View", _parentNamespace, viewModel, model, RelativePath);

			_classes.Add(backgroundClass);
			_views.Add(view);

			// view model : designer property
			ReadonlyPropertyDefinition designerProperty = new ReadonlyPropertyDefinition(viewModel, "Designer", true);
			viewModel.AddProperty(designerProperty);

			CreateAction createDomain = new CreateAction(sourceClass is CustomClass ?
				((CustomClass)sourceClass).DefaultConstructor : new ConstructorDefinition(sourceClass));
			designerProperty.AddCodeLineToGetter(createDomain);

			for (int index = 0; index < designerContent.Properties.Count; index++)
				AddInitializationActionsFor(designerProperty, createDomain.Result, designerContent.Properties[index]);

			CreateAction createModel = new CreateAction(contructorModel, createDomain.Result);
			designerProperty.AddCodeLineToGetter(createModel);

			CreateAction createViewModel = new CreateAction(contructorViewModel, createModel.Result);
			designerProperty.AddCodeLineToGetter(createViewModel);

			ReturnAction retirnViewModelForDesigner = new ReturnAction(createViewModel.Result);
			designerProperty.AddCodeLineToGetter(retirnViewModelForDesigner);

			return new WindowDefinition(model, viewModel, view);
		}

		public WindowContainer AddWindow(string name, CustomClass domainClass, ClassContent domainClassContent, List<PropertyPath> publicPropertiesForShellViewModel)
		{
			if (string.IsNullOrEmpty(name))
				throw new ArgumentNullException("name");

			string modelClassName = name.EndsWith("View") ? name.Remove(name.Length - 4, 4) + "Model" : name + "Model";
			string viewModelClassName = name.EndsWith("View") ? name + "Model" : name + "ViewModel";
			string viewName = name.EndsWith("View") ? name : name + "View";
			string backgroundClassName = viewName;

			// model
			CustomClass modelClass = new CustomClass(modelClassName, Namespace, false);
			MethodParameter domainClassAsParameter = new MethodParameter("domainClass", domainClass);
			ConstructorDefinition modelConstrucor = modelClass.AddConstructor(domainClassAsParameter);
        	Field backgroundField = modelConstrucor.GetBindedField(domainClassAsParameter);

			for (int index = 0; index < publicPropertiesForShellViewModel.Count; index++)
			{
				PropertyPath path = publicPropertiesForShellViewModel[index];
				StringBuilder builder = new StringBuilder();

				builder.Append(backgroundField.Name);
				path.ForEach(x => builder.AppendFormat(".{0}", x));

				modelClass.AddProperty(builder.ToString(), path.Name, path.Type);
			}

			// view model
			CustomClass viewModelClass = new CustomClass(viewModelClassName, Namespace, false, new StandartClass(typeof(ViewModelBase).Name, typeof(ViewModelBase).Namespace));
			MethodParameter modelClassAsParameter = new MethodParameter("model", modelClass);
			ConstructorDefinition viewModelConstrucor = viewModelClass.AddConstructor(modelClassAsParameter);
			backgroundField = viewModelConstrucor.GetBindedField(modelClassAsParameter);

			for (int index = 0; index < publicPropertiesForShellViewModel.Count; index++)
			{
				PropertyPath path = publicPropertiesForShellViewModel[index];
				StringBuilder builder = new StringBuilder();

				builder.Append(backgroundField.Name);
				builder.AppendFormat(".{0}", path.Name);

				viewModelClass.AddReadonlyProperty(builder.ToString(), path.Name, path.Type);
			}


			// view model : designer property
			ReadonlyPropertyDefinition designerProperty = new ReadonlyPropertyDefinition(viewModelClass, "Designer", true);
        	viewModelClass.AddProperty(designerProperty);

			CreateAction createDomain = new CreateAction(domainClass.Constructors.First());
			designerProperty.AddCodeLineToGetter(createDomain);

			for (int index = 0; index < domainClassContent.Properties.Count; index++)
				AddInitializationActionsFor(designerProperty, createDomain.Result, domainClassContent.Properties[index]);

			CreateAction createModel = new CreateAction(modelConstrucor, createDomain.Result);
			designerProperty.AddCodeLineToGetter(createModel);

			CreateAction createViewModel = new CreateAction(viewModelConstrucor, createModel.Result);
			designerProperty.AddCodeLineToGetter(createViewModel);

			ReturnAction retirnViewModelForDesigner = new ReturnAction(createViewModel.Result);
			designerProperty.AddCodeLineToGetter(retirnViewModelForDesigner);

			// window
			CustomClass backgroundClass = new CustomClass(backgroundClassName, Namespace, true, new StandartClass(typeof(Window).Name, typeof(Window).Namespace));
        	ConstructorDefinition constructor = backgroundClass.AddConstructor();
			constructor.AddCodeLine(new CallMethodAction(new ThisField(backgroundClass), new MethodDefinition("InitializeComponent")));

			WindowContainer viewClass = new WindowContainer(viewName, _parentNamespace, viewModelClass, modelClass, RelativePath);

			_classes.Add(modelClass);
			_classes.Add(viewModelClass);
			_classes.Add(backgroundClass);
			_views.Add(viewClass);


			return viewClass;
		}

		private static void AddInitializationActionsFor(ReadonlyPropertyDefinition content, LocalField initializedField, PropertyContent property)
		{
			if (property.Value != null)
			{
				PropertyDefinition targetProperty = (PropertyDefinition)((CustomClass)initializedField.Type).Properties.Find(x => x.Name == property.Name);
				SetPropertyValueAction setProperty = new SetPropertyValueAction(initializedField, targetProperty, new Field(property.Value, targetProperty.ReturnedType, false, false));

				content.AddCodeLineToGetter(setProperty);
			}
			else if (property.ArrayItems.Count > 0)
			{
				PropertyDefinition targetProperty = (PropertyDefinition)((CustomClass)initializedField.Type).Properties.Find(x => x.Name == property.Name);
				ClassBase templateTargetType = ((CollectionStandartClass) targetProperty.ReturnedType).TemplateType;

				if (targetProperty.ReturnedType is CollectionStandartClass)
				{
					ConstructorDefinition constructor = new ConstructorDefinition(targetProperty.ReturnedType);
					CreateAction createItem2 = new CreateAction(constructor);
					content.AddCodeLineToGetter(createItem2);

					SetPropertyValueAction setItem2 = new SetPropertyValueAction(initializedField, targetProperty, createItem2.Result);
					content.AddCodeLineToGetter(setItem2);
				}

				if (!(templateTargetType is CustomClass))
					return;

				for (int index = 0; index < property.ArrayItems.Count; index++)
				{
					PropertyContent itemContent = property.ArrayItems[index];

					CreateAction createItem = new CreateAction(((CustomClass) templateTargetType).DefaultConstructor);
					content.AddCodeLineToGetter(createItem);

					itemContent.Properties.ForEach(x => AddInitializationActionsFor(content, createItem.Result, x));

					CustomAction addProperty = new CustomAction("{0}.{1}.Add({2});", initializedField, targetProperty, createItem.Result);
					content.AddCodeLineToGetter(addProperty);
				}
			}
			else if (property.Properties.Count > 0)
			{
				ReadonlyPropertyDefinition targetProperty = ((CustomClass)initializedField.Type).Properties.Find(x => x.Name == property.Name);

				if (targetProperty.ReturnedType is CollectionStandartClass)
				{
					ConstructorDefinition constructor = new ConstructorDefinition(targetProperty.ReturnedType);
					CreateAction createItem2 = new CreateAction(constructor);
					content.AddCodeLineToGetter(createItem2);

					return;
				}

				CreateAction createItem = new CreateAction(((CustomClass)targetProperty.ReturnedType).DefaultConstructor);
				content.AddCodeLineToGetter(createItem);
				
				property.Properties.ForEach(x => AddInitializationActionsFor(content, createItem.Result, x));

				SetPropertyValueAction setProperty = new SetPropertyValueAction(initializedField, (PropertyDefinition)targetProperty, createItem.Result);
				content.AddCodeLineToGetter(setProperty);
			}
		}

		public virtual void Save(string directoryPath)
		{
			if (string.IsNullOrEmpty(directoryPath))
				throw new ArgumentNullException("directoryPath");

			if (!Directory.Exists(directoryPath))
				Directory.CreateDirectory(directoryPath);

			_classes.ForEach(classItem =>
			{
				string classPath = Path.Combine(directoryPath, classItem.NameWithExtension);
				classItem.Save(classPath);
			});

			_views.ForEach(viewItem =>
			{
				string classPath = Path.Combine(directoryPath, viewItem.Name + ViewBase.Extension);
				viewItem.Save(classPath);
			});

			_folders.ForEach(folderItem =>
			{
				string folderPath = Path.Combine(directoryPath, folderItem.Name);
				folderItem.Save(folderPath);
			});
		}

		public virtual ClassBase GetClass(string typeName)
		{
			CustomClass result = _classes.Find(x => x.Name == typeName);

			if (result != null)
				return result;

			for (int index = 0; index < _folders.Count; index++)
			{
				Folder child = _folders[index];
				ClassBase foundClass = child.GetClass(typeName);

				if (foundClass != null)
					return foundClass;
			}

			return null;
		}

		public ViewBase GetView(string name)
		{
			return GetAllViews().FirstOrDefault(x => x.Name == name);
		}

		internal ClassPosition[] GetAllCustomClasses()
		{
			return GetAllCustomClasses(string.Empty, string.Empty);
		}

		internal ClassPosition[] GetAllCustomClasses(string parentPath, string thisFolderPath)
		{
			List<ClassPosition> classes = new List<ClassPosition>();

			for (int index = 0, count = _classes.Count; index < count; index++)
			{
				CustomClass item = _classes[index];
				string itemPath = Path.Combine(thisFolderPath, item.NameWithExtension);

				ClassPosition position = new ClassPosition(item, itemPath);
				classes.Add(position);
			}

			classes.AddRange(_folders.SelectMany(x =>
			{
				string xPath = string.IsNullOrEmpty(thisFolderPath) ? x.Name : Path.Combine(thisFolderPath, x.Name);
				return x.GetAllCustomClasses(thisFolderPath, xPath);
			}));

			return classes.ToArray();
		}

		public ViewBase[] GetAllViews()
		{
			List<ViewBase> views = new List<ViewBase>();

			views.AddRange(_views);
			views.AddRange(_folders.SelectMany(x => x.GetAllViews()));

			return views.ToArray();
		}
	}

	public class ViewDefinition
	{
		private readonly CustomClass _model;
		private readonly CustomClass _viewModel;
		private readonly ViewContainer _view;

		public ViewDefinition(CustomClass model, CustomClass viewModel, ViewContainer view)
		{
			_model = model;
			_viewModel = viewModel;
			_view = view;
		}

		public CustomClass Model
		{
			get { return _model; }
		}

		public CustomClass ViewModel
		{
			get { return _viewModel; }
		}

		public ViewContainer View
		{
			get { return _view; }
		}
	}

	public class WindowDefinition
	{
		private readonly CustomClass _model;
		private readonly CustomClass _viewModel;
		private readonly WindowContainer _view;

		public WindowDefinition(CustomClass model, CustomClass viewModel, WindowContainer view)
		{
			_model = model;
			_viewModel = viewModel;
			_view = view;
		}

		public CustomClass Model
		{
			get { return _model; }
		}

		public CustomClass ViewModel
		{
			get { return _viewModel; }
		}

		public WindowContainer View
		{
			get { return _view; }
		}
	}

	public class ClassViewHierarchy
	{
		private readonly ClassBase _target;
		private readonly string _propertyName;
		private readonly List<ClassViewProperty> _viewProperties = new List<ClassViewProperty>();

		public ClassViewHierarchy(ClassBase target, string propertyName)
		{
			_target = target;
			_propertyName = propertyName;
		}

		public string PropertyName
		{
			get { return _propertyName; }
		}

		public ClassBase Target
		{
			get { return _target; }
		}

		public List<ClassViewProperty> ViewProperties
		{
			get { return _viewProperties; }
		}

		public override string ToString()
		{
			return string.Format("Type: {0}, Children count: {1}", _target, ViewProperties.Count);
		}
	}

	public class ClassViewProperty
	{
		private readonly List<ReadonlyPropertyDefinition> _path = new List<ReadonlyPropertyDefinition>();
		private readonly ClassViewHierarchy _target;

		public ClassViewProperty(ClassViewHierarchy target)
		{
			_target = target;
		}

		public ClassViewHierarchy Target
		{
			get { return _target; }
		}

		public List<ReadonlyPropertyDefinition> Path
		{
			get { return _path; }
		}

		public override string ToString()
		{
			return _target.ToString();
		}
	}
}
